/* $Id: asm.c,v 1.8 2000/11/27 20:09:43 malekith Exp $ */
#include "compiler.h"
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

extern char *op_codes[];
extern int debug;

#ifdef NO_ALIGN
#define align(ctx) do { } while (0)
#else
static void align(CTX *ctx)
{
	int n;

	n = (ALIGN - ((db_off(ctx->code) + 1) % ALIGN)) % ALIGN;
	
	if (debug)
		printf("ALIGN +%d bytes\n", n);
	while (n--)
		emit(ctx, op_nop, 0);
}
#endif

static void dump_bytes(void *ptr, int n)
{
	byte *p;

	p = (byte*)ptr;
	while (n--)
		printf(" 0x%02x", *p++);
	printf("\n");
}

static void branch(CTX *ctx, int op, u32 id)
{
	align(ctx);
	db_addc(ctx->code, op);
	add_u32(reloc, db_off(ctx->code));
	*(u32*)db_alloc(ctx->code, 4) = id;

	if (debug) {
		char xbuf[100];
		
		snprintf(xbuf, sizeof(xbuf) - 1, op_codes[op], id);
		printf("\t%-40s ; 0x%02x", xbuf, op);
		dump_bytes(&id, 4);
	}
}

static void load_string(CTX *ctx, const char *s)
{
	int id;

	id = add_id(ctx);
	RES[id] = db_off(ctx->strings);
	
	align(ctx);
	db_addc(ctx->code, op_mov_s);
	add_u32(reloc, db_off(ctx->code));
	*(u32*)db_alloc(ctx->code, 4) = id;

	if (debug) {
		char xbuf[100];
		
		snprintf(xbuf, sizeof(xbuf) - 1, op_codes[op_mov_s], s);
		printf("\t%-40s ; 0x%02x", xbuf, op_mov_s);
		dump_bytes(&id, 4);
	}
	
	add_u32(strings_ids, id);
	
	for (s++; *s != '"'; s++)
		if (*s == '\\')
			switch (*++s) {
			case 'n':
				db_addc(ctx->strings, '\n');
				break;
			case 'e':
				db_addc(ctx->strings, '\n');
				break;
			case 'b':
				db_addc(ctx->strings, '\b');
				break;
			case 't':
				db_addc(ctx->strings, '\t');
				break;
			default:
				db_addc(ctx->strings, *s);
				break;
			}
		else
			db_addc(ctx->strings, *s);
			
	db_addc(ctx->strings, 0);
}

static void load_float(CTX *ctx, real d)
{
	if (d == 0.0) {
		emit(ctx, op_clr_f, 0);
		return;
	}

	align(ctx);
	db_addc(ctx->code, op_mov_f);
	*(real*)db_alloc(ctx->code, sizeof(real)) = d;
	
	if (debug) {
		char xbuf[100];
		
		snprintf(xbuf, sizeof(xbuf) - 1, op_codes[op_mov_f], d);
		printf("\t%-40s ; 0x%02x", xbuf, op_mov_f);
		dump_bytes(&d, sizeof(d));
	}
}

static void load_long(CTX *ctx, iptr d)
{
	iptr a;
	char xbuf[100];

	a = (d < 0) ? -d : d;

	/* FIXME: long long support */
	if (a >= 0x7fff) {
		align(ctx);
		db_addc(ctx->code, op_mov_l);
		*(u32*)db_alloc(ctx->code, sizeof(u32)) = d;

		if (debug) {
			u32 tmp;
			
			tmp = (u32)d;
			snprintf(xbuf, sizeof(xbuf) - 1, 
				 op_codes[op_mov_l], d);
			printf("\t%-40s ; 0x%02x", xbuf, op_mov_l);
			dump_bytes(&tmp, 4);
		}
	} else if (a >= 0x7f) {
		align(ctx);
		db_addc(ctx->code, op_mov_w);
		*(u16*)db_alloc(ctx->code, sizeof(u16)) = (u16)(d & 0xffff);
		
		if (debug) {
			u16 tmp;
			
			tmp = (u16)d;
			snprintf(xbuf, sizeof(xbuf) - 1, 
				 op_codes[op_mov_w], d);
			printf("\t%-40s ; 0x%02x", xbuf, op_mov_w);
			dump_bytes(&tmp, 2);
		}
	} else if (a > 0) {
		db_addc(ctx->code, op_mov_b);
		db_addc(ctx->code, d & 0xff);
		if (debug) {
			snprintf(xbuf, sizeof(xbuf) - 1, 
				 op_codes[op_mov_b], d);
			printf("\t%-40s ; 0x%02x 0x%02x\n", xbuf, op_mov_b, 
			       d & 0xff);
		}
	} else
		emit(ctx, op_clr_d, 0);
}

static void class_ref(CTX *ctx, int code, struct class_def *c)
{
	align(ctx);
	db_addc(ctx->code, code);
	add_u32(class_xrefs, db_off(ctx->code));
	*(u32*)db_alloc(ctx->code, sizeof(u32)) = c->id;
	
	if (debug) {
		char xbuf[100];
		u32 x;

		x = c->id;
		snprintf(xbuf, sizeof(xbuf) - 1, op_codes[code], c->name);
		printf("\t%-40s ; 0x%02x", xbuf, code);
		dump_bytes(&x, sizeof(x));
	}
}

void emit(CTX *ctx, ...)
{
	va_list ap;
	int code;


	/* FIXME: add stack counters/checkers */
	
	if (ctx->dont_emit)
		return;

	va_start(ap, ctx);

	while ((code = va_arg(ap, int))) 
		switch (code) {
		case op_mov_f:
			load_float(ctx, va_arg(ap, double));
			break;
		case op_mov_l:
			load_long(ctx, va_arg(ap, iptr));
			break;
		case op_mov_s:
			load_string(ctx, va_arg(ap, char*));
			break;
		case op_bra:
		case op_btrue:
		case op_bfalse:
		case op_frame:	/* does the same ... */
			branch(ctx, code, va_arg(ap, int));
			break;
		case op_new:
		case op_alloc:
		case op_jsr_id:
			class_ref(ctx, code, va_arg(ap, struct class_def*));
			break;
		default:
			if (code >= op_last)
				halt();
			if (debug)
				printf("\t%-40s ; 0x%02x\n", 
					op_codes[code], code);
			db_addc(ctx->code, code);
			break;
		}
		
	va_end(ap);
}
