/* $Id: token.c,v 1.7 2000/11/27 20:09:44 malekith Exp $ */

#include "util.h"
#include "token.h"
#include "y.tab.h"
#include "phash.h"
#include <stdlib.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>

struct token *root;
		
extern int yydebug;
int yyparse(void);

int compile_class(struct token *t);
int compile_func(struct token *t);
int write_out();

extern int debug;

void cpp_init(const char *fn);
int cpp_get();
extern char *cpp_file;
extern int cpp_line;
char *in_file;
int stats;

/* this simple function saves us 1/4 of RSS and 1/6 of cpu ...
 * note that you cannot free that memory ! */
void *my_malloc(int need)
{
	static byte *chunk;
	static int size, pos;

	if (chunk == 0) {
		pos = 0;
		chunk = malloc(size = 1024 * 16);
	}

#ifndef NO_ALIGN
	if (need & (ALIGN - 1))
		need += ALIGN - (need & (ALIGN - 1));
#endif

	if (need <= size - pos) {
		pos += need;	
		return chunk + (pos - need);
	}
	if (need > size)
		return malloc(need);
		
	chunk = 0;
	return my_malloc(need);
}

char *my_strdup(const char *s)
{
	return strcpy(my_malloc(strlen(s) + 1), s);
}

struct token *new_token(struct token *left, struct token *right, int sel, ... )
{
	struct token *r;
	va_list ap;

	/* 
	 * FIXME: force this to use one (or few) memory segment 
	 * for all objects. that segment could be freed as a whole later
	 * on...
	 */
	r = (struct token *)my_malloc(sizeof(struct token));
	r->left = left;
	r->right = right;
	r->line = cpp_line;
	r->file = cpp_file;
	va_start(ap, sel);
	switch (r->sel = sel) {
	case sel_kw:
	case sel_op:
	case sel_int:
	case sel_syn:
		r->v_int = va_arg(ap, int);
		break;
	case sel_ptr:
		r->v_ptr = 0;
		break;
	case sel_double:
		r->v_double = va_arg(ap, double);
		break;
	case sel_id:
	case sel_string:
		r->v_string = my_strdup(va_arg(ap, char*));
		break;
	default:
		halt();
	}
	return r;
}

void yyerror(const char *s)
{
	fprintf(stderr, "%s:%d: %s\n", cpp_file, cpp_line, s);
}

void oops(struct token *t, const char *fmt, ...)
{
	va_list ap;
	va_start(ap, fmt);
	fprintf(stderr, "%s:%d: ", t->file, t->line);
	vfprintf(stderr, fmt, ap);
	fprintf(stderr, "\n");
}

int compile(struct token *t)
{
	static int lev;
	int r = 1;
	
	lev++;
	if (debug)
		printf("# lev = %d\n", lev);

restart:

	if (t == 0)
		goto out;
/* FIXME: use bounded stack! */
	assert(t->sel == sel_syn);
	switch (t->v_int) {
	case syn_file:
		if (compile(t->left)) 
			goto fail;
		t = t->right;
		goto restart;
	case syn_class:
		if (compile_class(t))
			goto fail;
		break;
	case syn_func:
		if (compile_func(t))
			goto fail;
		break;
	default:	
		halt();
	}

out:
	r = 0;
fail:
	lev--;
	return r;
}

void dump_stats(char *s)
{
	FILE *f;
	int sys=0,us=0,vm=0,rss=0;

	if (!stats)
		return;
		
	f = fopen("/proc/self/stat", "r");
	if (f == 0)
		return;
	fscanf(f, "%*d %*s %*c %*d %*d %*d %*d %*d %*d %*d %*d %*d %*d "
		"%d %d %*d %*d %*d %*d %*d %*d %*d %d %d",
		&us,&sys,&vm,&rss);
	printf("%s: sys: %d, user: %d [%d]   vm: %d rss: %d\n",s,
		sys,us,sys+us,vm/1024,rss*4);
	
}

int main(int argc, char **argv)
{
	FILE *f;
	char *name;
	int cpp_only = 0, c;
	char *p;
	
	(void)argc;
	name = *argv++;
	while (*argv && **argv == '-') {
		for (p = argv[0] + 1; *p; p++)
			switch (*p) {
			case 'y':
				yydebug = 1;
				break;
			case 'd':
				debug = 1;
				break;
			case 'E':
				cpp_only = 1;
				break;
			case 's':
				stats = 1;
				break;
			default:
				fprintf(stderr, "Unknown option: -%c\n", *p);
				exit(1);
			}
		argv++;
	}
	
	if (*argv) {
		f = fopen(*argv, "r");
		if (f == 0) {
			perror(*argv);
			exit(1);
		}
		fclose(f);
		in_file = *argv;
	} else {
		fprintf(stderr, "USAGE: [-ydsE] %s filename.k\n", name);
		exit(1);
	}
	
	cpp_init(in_file);
	
	if (cpp_only) {
		while ((c = cpp_get()) != -1)
			putchar(c);
		putchar('\n');
		exit(0);
	}
	
	dump_stats("start");
	if (yyparse())
		return 1;
	dump_stats("yyparse");
	if (compile(root))
		return 1;
	dump_stats("compile");
	if (write_out(root))
		return 1;
	dump_stats("write_out");

	destroy_token(root);
	return 0;
}

void destroy_token(struct token *t)
{
	(void)t;
}
